import os

import globals
import lexer
import utils

_running_id_number = 1

_EXPR_TYPE_DICT = {
    globals.E_BOOLEAN: 'boolean',
    globals.E_INT: 'int',
    globals.E_FLOAT: 'float',
    globals.E_STRING: 'string',
    globals.E_FILENAME: 'filename',
    globals.E_DATETIME: 'datetime',

    globals.E_STRING_FILENAME_DATETIME: '<string or filename or datetime>',
}

def type2str(expr_type):
    if expr_type is None:
        return '<UNKNOWN>'
    if expr_type not in _EXPR_TYPE_DICT:
        raise ValueError, 'Unknown expression type: %s' % expr_type
    return _EXPR_TYPE_DICT[expr_type]


class ASTNode:
    def __init__(self, parent, rootdir):
        if not rootdir:
            raise ValueError, 'Rootdir not set'

        assert '//' not in rootdir
        assert '\\' not in rootdir
        self.parent = parent
        self.rootdir = utils.abspath(rootdir)
        self.token = None

        global _running_id_number
        self.idnumber = _running_id_number
        _running_id_number += 1

    def visit(self, visitor, arg=None):
        class_name = self.__class__.__name__
        method = getattr(visitor, 'visit_' + class_name)
        return method(self, arg)

    def set_rootdir(self, rootdir):
        self.rootdir = utils.abspath(rootdir)

    def get_parents(self):
        """Return list of parent nodes in order from direct parent
        to root node."""
        parents = []
        cur_parent = self.parent
        while cur_parent:
            parents.append(cur_parent)
            cur_parent = cur_parent.parent
        return parents

    def under_in_block(self):
        """Return whether node is an IN-block or a (grand)child of an IN-block."""
        for parent in self.get_parents():
            if parent.rootdir != self.rootdir:
                return True
        return False

### Rules ##############################################################

class Rule(ASTNode):
    def __init__(self, parent, rootdir):
        ASTNode.__init__(self, parent, rootdir)

    def innerdir_common_prefix(self):
        """Return the common prefix of all innerdirs of the children
        of the node."""
        raise NotImplementedError

class BlockRule(Rule):
    def __init__(self, parent, rootdir):
        Rule.__init__(self, parent, rootdir)
        self.rules = []

    def __str__(self):
        if self.parent:
            parentstr = '%d' % self.parent.idnumber
        else:
            parentstr= '-'
        return 'Block rule (length %d, id %s, parent %s, rootdir "%s")' % (len(self.rules), self.idnumber, parentstr, self.rootdir)

    def innerdir_common_prefix(self):
        children_prefixes = [c.innerdir_common_prefix() for c in self.rules]
        return os.path.commonprefix(children_prefixes)


class IfBlockRule(BlockRule):
    def __init__(self, parent, rootdir):
        BlockRule.__init__(self, parent, rootdir)
        self.if_expr = None

    def __str__(self):
        if self.parent:
            parentstr = '%d' % self.parent.idnumber
        else:
            parentstr= '-'
        return 'IF-block rule (length %d, id %s, parent %s, rootdir "%s")' % (len(self.rules), self.idnumber, parentstr, self.rootdir)


class SelectorRule(Rule):
    def __init__(self, parent, rootdir, inclusive):
        Rule.__init__(self, parent, rootdir)
        self.inclusive = inclusive


class GloblistRule(SelectorRule):
    def __init__(self, parent, rootdir, inclusive):
        SelectorRule.__init__(self, parent, rootdir, inclusive)
        self.glob_patterns = []
        self.if_expr = None

    def __str__(self):
        if self.parent:
            parentstr = '%d' % self.parent.idnumber
        else:
            parentstr= '-'
        if self.inclusive:
            exclstr = ''
        else:
            exclstr = 'EXCLUDE '
        return 'Glob list %s(length %d, parent %s, rootdir "%s")' % (exclstr, len(self.glob_patterns), parentstr, self.rootdir)

    def innerdir_common_prefix(self):
        children_prefixes = [c.innerdir for c in self.glob_patterns]
        return os.path.commonprefix(children_prefixes)


class ForEachRule(SelectorRule):
    def __init__(self, parent, rootdir, inclusive):
        SelectorRule.__init__(self, parent, rootdir, inclusive)
        self.variable_expr = None
        self.glob_patterns = []
        self.if_expr = None

    def __str__(self):
        if self.inclusive:
            exclstr = ''
        else:
            exclstr = 'EXCLUDE '
        return 'For each %s(variable "%s", %s globs, rootdir "%s")' % (exclstr, self.variable_expr.variable_name, len(self.glob_patterns), self.rootdir)

    def innerdir_common_prefix(self):
        children_prefixes = [c.innerdir for c in self.glob_patterns]
        return os.path.commonprefix(children_prefixes)


class GlobPattern(ASTNode):
    def __init__(self, parent, rootdir, pattern, recursive):
        ASTNode.__init__(self, parent, rootdir)
        self.pattern = pattern.replace('\\', '/')
        self.recursive = recursive
        self.is_absolute = utils.filename_is_absolute(self.pattern)

        if not self.is_absolute:
            self.fullglob, self.innerdir = utils.calc_fullglob_innerdir(self.rootdir, self.pattern)
        else:
            self.fullglob = utils.abspath(self.pattern)
            self.innerdir = utils.calc_innerdir(self.fullglob)

    def __str__(self):
        if not self.recursive:
            return 'Glob: "%s" NONREC (innerdir "%s")' % (self.fullglob, self.innerdir)
        else:
            return 'Glob: "%s" (innerdir "%s")' % (self.fullglob, self.innerdir)

### Expressions ########################################################

class Expression(ASTNode):
    def __init__(self, parent, rootdir, expr_type):
        ASTNode.__init__(self, parent, rootdir)
        self.expr_type = expr_type


class AndExpression(Expression):
    def __init__(self, parent, rootdir):
        Expression.__init__(self, parent, rootdir, globals.E_BOOLEAN)
        self.expr1 = None
        self.expr2 = None

    def __str__(self):
        return 'AND expression'


class OrExpression(Expression):
    def __init__(self, parent, rootdir):
        Expression.__init__(self, parent, rootdir, globals.E_BOOLEAN)
        self.expr1 = None
        self.expr2 = None

    def __str__(self):
        return 'OR expression'


class NotExpression(Expression):
    def __init__(self, parent, rootdir,):
        Expression.__init__(self, parent, rootdir, globals.E_BOOLEAN)
        self.expr = None

    def __str__(self):
        return 'NOT expression'


class FunctionExpression(Expression):
    def __init__(self, parent, rootdir, funcname):
        Expression.__init__(self, parent, rootdir, None)
        self.funcname = funcname
        self.params = [] # [Expression]

    def __str__(self):
#        paramstr = ', '.join([str(p) for p in self.params])
        return 'Function call (funcname "%s", %d params)' % (self.funcname, len(self.params))


class CompareExpression(Expression):
    def __init__(self, parent, rootdir, expr1, operator_type):
        Expression.__init__(self, parent, rootdir, globals.E_BOOLEAN)
        self.expr1 = expr1
        self.expr2 = None
        self.operator_type = operator_type

    def __str__(self):
        s = lexer.type2str(self.operator_type)
        return 'Compare expression (type "%s")' % s


class LiteralExpression(Expression):
    def __init__(self, parent, rootdir, strvalue, atom_type):
        Expression.__init__(self, parent, rootdir, atom_type)
        self.strvalue = strvalue
        self.datetime_value = None
        self.filename_value = None

    def __str__(self):
        if self.expr_type in (globals.E_STRING, globals.E_STRING_FILENAME_DATETIME):
            s = '"%s"' % self.strvalue
        elif self.expr_type == globals.E_FILENAME:
            s = self.filename_value
        elif self.expr_type == globals.E_DATETIME:
            s = self.datetime_value.strftime('%Y-%m-%d %H:%M:%S')
        else:
            s = self.strvalue
        return 'Atom (%s, type %s)' % (s, type2str(self.expr_type))


class VariableExpression(Expression):
    def __init__(self, parent, rootdir, variable_name):
        Expression.__init__(self, parent, rootdir, None)
        self.variable_name = variable_name
        self.variable_obj = None # Reference to symboltable.Variable

    def __str__(self):
        return 'Variable (%s, type %s)' % (self.variable_name, type2str(self.expr_type))
