import cStringIO
import os
import sys
import unittest

from fsl import functions, Interpreter, fslparser
import fsl.utils

import TesterHookHandler
import tutils

# set and frozenset were added in Python 2.4, but
# we want to support 2.3, too.
# Well, perhaps we should just use sets.Set and
# sets.ImmutableSet everywhere...
try:
    set, frozenset
except NameError:
    import sets
    set = sets.Set
    frozenset = sets.ImmutableSet

"""
Source file structures to test
------------------------------

Note:
- all exclusion rule tests (with NOT) are done in combination with
  another (inclusive) rule
- all tests should be performed for these rootdirs (6):
  * rootdir
  * rootdir/
  * /abs/path/to/rootdir
  * /abs/path/to/rootdir/
  * ./rootdir
  * ./rootdir/

1. Single glob pattern
- pattern is directory name | pattern is file name | pattern is * | pattern has * | pattern has ?
- pattern begins with "/" | pattern has "/" | has not "/"
- without NOT | with NOT
- without NONREC | with NONREC
- pattern matches file(s) | doesn't match file(s)
* special: pattern = "and"
Total: 5*3*2*2*2 + 1 = 121 combinations

2. Glob list, 2 glob patterns
- glob list without NOT | with NOT
- pattern1 without NONREC | without NONREC
- pattern2 without NONREC | without NONREC
Total: 2*2*2 = 8 combinations

3. Glob list, 4 glob patterns
- glob list without NOT | with NOT
Total: 2 combinations

Now, we assume globs and glob lists work. We still test glob list rules
and for-each rules with 1 and 2 glob pattens.

4. Globlist rule with IF-expression
- 1 glob pattern | 2 glob patterns
- glob list without NOT | with NOT
- IF-expression is false | IF-expression is true | IF-expression is invalid
Total: 2*2*3 = 12 combinations

5. For-each rule
- 1 glob pattern | 2 glob patterns
- without NOT | with NOT
- IF-expression true for all files | false for all files | true for some files | invalid
Total: 2*2*4 = 16 combinations

6. Several rules in one file, all rules inclusive
* 2 glob list rules
* 2 for each rules
* 2 IN-blocks
* glob list, for-each rule, IN-block
* glob list, IN-block, for-each rule
* for-each, glob list, IN-block
* for-each, IN-block, glob list
* IN-block, glob list rule, for-each
* IN-block, for-each, glob list rule
* 8 rules in sequence
Total: 10 combinations

Now, we assume rule sequences work.

7. IN-blocks
- existing directory | nonexisting directory
- 1 glob list rule | 1 for-each rule | 3 glob list rules
Total: 2*3 = 6 combination

8. Nested IN-blocks
* IN dir1
      IN dir2
          glob list rule
* IN dir1
      glob list rule
      IN dir2
          glob list rule
      glob list rule
* IN dir1
      IN dir2
          IN dir3
              glob list rule
* IN dir1
      IN dir2
          IN dir3
              glob list rule
      glob list rule
Total: 4

9. Inclusion/exclusion combinations
* include all, exclude all                                -> result {}
* exclude all, include all                                -> result *
* include all, exclude all, include all                   -> result *
* include all, exclude all, include all, exclude all      -> result {}
* include {A B C}, exclude {A}                            -> result {B C}
* include {A B C}, exclude {A}, exclude {A}               -> result {B C}
* include {A B C}, exclude {B}, exclude {C}               -> result {A}
* include {A B C}, exclude {B}, exclude {A}, exclude {C}  -> result {}
* include {A B C}, exclude {A}, include {A}               -> result {A B C}
* include {A B C}, exclude {A}, include {D}               -> result {B C D}
Total: 10

10. Expressions
* NOT: expression true | expression false
* AND: true + true | true + false | false + true | false + false
* OR: true + true | true + false | false + true | false + false
* Compare: < <= > >= = != (6)
  - types: int & int | int & float | float & int | string & string |
    datetime & datetime
* Functions: now(), size(file), extract(date, part)
  - size(file): parameter is literal | parameter is variable

Grand total: 81 + 8 + 2 + 12 + 16 + 10 + 6 + 4 + 10 = 189
"""

ALL_FILES = frozenset([
    "and",
    "filea",
    "fileb.txt",
    "filec.txt.exe",
    "filed",
    "ALLCAPS",
    ".hidden",
    "long-filename-1234567890-abcdefghijkl",
    "dir1/filea",
    "dir1/filea-a-a",
    "dir1/dir2/asdf",
    "dir1/dir2/dir3/filez",
    "dir1/dir2/dir3/qwerty",
    "dir xyz/aaa",
    "dir xyz/filea",
    "dir xyz/q q q",
])

FIRST_LEVEL_FILES = frozenset([
    "and",
    "filea",
    "fileb.txt",
    "filec.txt.exe",
    "filed",
    "ALLCAPS",
    ".hidden",
    "long-filename-1234567890-abcdefghijkl",
])

# file* NONREC
FIRST_LEVEL_FILE_ASTERISK = frozenset([
    "filea",
    "fileb.txt",
    "filec.txt.exe",
    "filed",
])

# */file*
ALL_FILE_ASTERISK = frozenset([
    "filea",
    "fileb.txt",
    "filec.txt.exe",
    "filed",
    "dir1/filea",
    "dir1/filea-a-a",
    "dir1/dir2/dir3/filez",
    "dir xyz/filea",
])

# *file*
ALL_ASTERISK_FILE_ASTERISK = frozenset([
    "filea",
    "fileb.txt",
    "filec.txt.exe",
    "filed",
    "long-filename-1234567890-abcdefghijkl",
    "dir1/filea",
    "dir1/filea-a-a",
    "dir1/dir2/dir3/filez",
    "dir xyz/filea",
])

DIR1_FILES = frozenset([
    "dir1/filea",
    "dir1/filea-a-a",
    "dir1/dir2/asdf",
    "dir1/dir2/dir3/filez",
    "dir1/dir2/dir3/qwerty",
])

DIR1_NONREC_FILES = frozenset([
    "dir1/filea",
    "dir1/filea-a-a",
])

# dir1/file*
DIR1_FILE_ASTERISK = frozenset([
    "dir1/filea",
    "dir1/filea-a-a",
])

DIR1_DIR2_FILES = frozenset([
    "dir1/dir2/asdf",
    "dir1/dir2/dir3/filez",
    "dir1/dir2/dir3/qwerty",
])

DIR1_DIR2_NONREC_FILES = frozenset([
    "dir1/dir2/asdf",
])

DIR1_DIR2_DIR3_FILES = frozenset([
    "dir1/dir2/dir3/filez",
    "dir1/dir2/dir3/qwerty",
])

TMP_SOURCE_FILE = 'tmpsrc'


SINGLE_GLOB_TESTS = [
    # Directory name
    ('dir1', DIR1_FILES),
    ('dir1 NONREC', DIR1_NONREC_FILES),
    ('dir1/dir2', DIR1_DIR2_FILES),
    ('dir1/dir2 NONREC', DIR1_DIR2_NONREC_FILES),

    # Filename
    ('filea',                      ['filea']),
    ('filea NONREC',               ['filea']),
    ('dir1/filea-a-a',             ['dir1/filea-a-a']),
    ('dir1/filea-a-a NONREC',      ['dir1/filea-a-a']),
    ('dir1/filea-a-a/',            ['dir1/filea-a-a']),
    ('dir1/filea-a-a/ NONREC',     ['dir1/filea-a-a']),

    # Filename, nonexisting
    ('notfound', []),
    ('notfound NONREC', []),
    ('dir1/notfound', []),
    ('dir1/notfound NONREC', []),

    # Pattern = *
    ('*',             ALL_FILES),
    ('* NONREC',      FIRST_LEVEL_FILES),

    # Pattern has one *
    ('file*', ['filea', 'fileb.txt', 'filec.txt.exe', 'filed']),
    ('file* NONREC', ['filea', 'fileb.txt', 'filec.txt.exe', 'filed']),
    ('dir1/f*', ['dir1/filea', 'dir1/filea-a-a']),
    ('dir1/fil* NONREC', ['dir1/filea', 'dir1/filea-a-a']),
    ('dir1*', DIR1_FILES),

    # Pattern has one *, nonexisting
    ('notfound*', []),
    ('notfound* NONREC', []),
    ('dir1/notfound*', []),
    ('dir1/notfound* NONREC', []),

    # Complex * patterns
    ('*/filea',             ["filea", "dir1/filea", "dir xyz/filea"]),
    ('*/filea*',            ["filea", "dir1/filea", "dir1/filea-a-a", "dir xyz/filea"]),
    ('*/file*',             ALL_FILE_ASTERISK),
    ('*/file* NONREC',      FIRST_LEVEL_FILE_ASTERISK),
    ('*file*',              ALL_ASTERISK_FILE_ASTERISK),
    ('*/*.*.*',             ["filec.txt.exe"]),  # Two extensions

    ('**',                  ALL_FILES),
    ('***',                 ALL_FILES),
    ('*/*',                 ALL_FILES),
    ('*/*/*/*',             DIR1_DIR2_FILES | DIR1_DIR2_DIR3_FILES),
    ('*/*/*/*/*/*',         []),
    ('.*',                  [".hidden"]),
    ('"* *filea"',          ["dir xyz/filea"]),
    ('*filea?',             []),

    # Pattern has ?
    ('file?',                     ['filea', 'filed']),
    ('file? NONREC',              ['filea', 'filed']),
    ('dir1/?ilea',                ['dir1/filea']),
    ('dir1/fi?ea NONREC',         ['dir1/filea']),

    # Pattern has ?, nonexisting
    ('?',                         []),
    ('notfound?',                 []),
    ('notfound? NONREC',          []),
    ('dir1/notfound?',            []),
    ('dir1/notfound? NONREC',     []),

    # Pattern = "and"
    ('"and"',            ['and']),
    ('"and" NONREC',     ['and']),
]

SINGLE_GLOB_EXCLUSION_TESTS = [
    ('*\nNOT dir1', ALL_FILES - DIR1_FILES),
    ('*\nNOT dir1 NONREC', ALL_FILES - DIR1_FILES),
    ('*\nNOT dir1/dir2', ALL_FILES - DIR1_DIR2_FILES),
    ('*\nNOT dir1/dir2 NONREC', ALL_FILES - DIR1_DIR2_FILES),
    ('*\nNOT dir2', ALL_FILES),

    ('*\nNOT filea', ALL_FILES - set(["filea"])),
    ('*\nNOT filea NONREC', ALL_FILES - set(["filea"])),
    ('*\nNOT dir1/file-a-a', ALL_FILES - set(["dir1/filea-a"])),
    ('*\nNOT dir1/filea-a-a NONREC', ALL_FILES - set(["dir1/filea-a-a"])),
    ('*\nNOT notfound', ALL_FILES),
    ('*\nNOT notfound NONREC', ALL_FILES),
    ('*\nNOT dir1/notfound', ALL_FILES),
    ('*\nNOT dir1/notfound NONREC', ALL_FILES),

    ('*\nNOT *', []),
    ('*\nNOT * NONREC', []),

    ('*\nNOT file*', ALL_FILES - FIRST_LEVEL_FILE_ASTERISK),
    ('*\nNOT file* NONREC', ALL_FILES - FIRST_LEVEL_FILE_ASTERISK),
    ('*\nNOT dir1/f*', ALL_FILES - set(["dir1/filea", "dir1/filea-a-a"])),
    ('*\nNOT dir1/fil* NONREC', ALL_FILES - set(["dir1/filea", "dir1/filea-a-a"])),

    ('*\nNOT notfound*', ALL_FILES),
    ('*\nNOT notfound* NONREC', ALL_FILES),
    ('*\nNOT dir1/notfound*', ALL_FILES),
    ('*\nNOT dir1/notfound* NONREC', ALL_FILES),

    ('*\nNOT */file*', ALL_FILES - ALL_FILE_ASTERISK),
    ('*\nNOT */file* NONREC', ALL_FILES - FIRST_LEVEL_FILE_ASTERISK),
    ('*\nNOT dir?', ALL_FILES - DIR1_FILES),

    ('*\nNOT file?', ALL_FILES - set(['filea', 'filed'])),
    ('*\nNOT file? NONREC', ALL_FILES - set(['filea', 'filed'])),
    ('*\nNOT dir1/?ilea', ALL_FILES - set(['dir1/filea'])),
    ('*\nNOT dir1/fi?ea NONREC', ALL_FILES - set(['dir1/filea'])),

    ('*\nNOT notfound?', ALL_FILES),
    ('*\nNOT notfound? NONREC', ALL_FILES),
    ('*\nNOT dir1/notfound?', ALL_FILES),
    ('*\nNOT dir1/notfound? NONREC', ALL_FILES),

    ('*\nNOT "and"', ALL_FILES - set(["and"])),
    ('*\nNOT "and" NONREC', ALL_FILES - set(["and"])),
]

# Source string, expected files

EXPRESSION_TESTS = [
    ('filea IF NOT 1=0',         ['filea']),
    ('filea IF NOT 1=1',         []),
    ('filea IF NOT NOT 1=0',     []),
    ('filea IF NOT NOT 1=1',     ['filea']),

    ('filea IF 1=1 AND 1=1',     ['filea']),
    ('filea IF 1=1 AND 1=0',     []),
    ('filea IF 1=0 AND 1=1',     []),
    ('filea IF 1=0 AND 1=0',     []),

    ('filea IF 1=1 OR 1=1',      ['filea']),
    ('filea IF 1=1 OR 1=0',      ['filea']),
    ('filea IF 1=0 OR 1=1',      ['filea']),
    ('filea IF 1=0 OR 1=0',      []),

    ('filea IF 1 < 3',           ['filea']),
    ('filea IF 3 < 1',           []),
    ('filea IF 1.5 < 3.5',       ['filea']),
    ('filea IF 3.5 < 1.5',       []),
    ('filea IF 1   < 3.5',       ['filea']),
    ('filea IF 3   < 1.5',       []),
    ('filea IF 1.5 < 3',         ['filea']),
    ('filea IF 3.5 < 1',         []),
    ('filea IF "abc" < "def"',   ['filea']),
    ('filea IF "def" < "abc"',   []),
    ('filea IF date("filea") < "2005-08-01 00:00:00"',   ['filea']),
    ('filea IF "2005-08-01 00:00:00" < date("filea")',   []),
    ('EACH f IN filea IF f < "xyz"',                     ['filea']),
    ('EACH f IN filea IF f < "abc"',                     []),
    ('EACH f IN filea IF "xyz" < f',                     []),

    ('filea IF 1 <= 3',           ['filea']),
    ('filea IF 1 <= 1',           ['filea']),
    ('filea IF 3 <= 1',           []),
    ('filea IF "abc" <= "def"',   ['filea']),
    ('filea IF "abc" <= "abc"',   ['filea']),
    ('filea IF "def" <= "abc"',   []),

    ('filea IF 3 > 1',           ['filea']),
    ('filea IF 1 > 3',           []),
    ('filea IF 3.5 > 1.5',       ['filea']),
    ('filea IF 1.5 > 3.5',       []),
    ('filea IF "abc" > "def"',   []),
    ('filea IF "def" > "abc"',   ['filea']),
    ('filea IF date("filea") > "2005-08-01 00:00:00"',   []),
    ('filea IF "2005-08-01 00:00:00" > date("filea")',   ['filea']),
    ('EACH f IN filea IF f > "xyz"',                     []),
    ('EACH f IN filea IF f > "abc"',                     ['filea']),
    ('EACH f IN filea IF "xyz" > f',                     ['filea']),

    ('filea IF 3 >= 1',           ['filea']),
    ('filea IF 1 >= 1',           ['filea']),
    ('filea IF 1 >= 3',           []),
    ('filea IF "def" >= "abf"',   ['filea']),
    ('filea IF "abc" >= "abc"',   ['filea']),
    ('filea IF "abc" >= "def"',   []),

    ('filea IF 1 = 1',           ['filea']),
    ('filea IF 3 = 1',           []),
    ('filea IF 1.0 = 1.0',       ['filea']),
    ('filea IF 3.0 = 1.0',       []),
    ('filea IF 1   = 1.0',       ['filea']),
    ('filea IF 3   = 1.0',       []),
    ('filea IF 1.0 = 1',         ['filea']),
    ('filea IF 3.0 = 1',         []),
    ('filea IF "abc" = "abc"',   ['filea']),
    ('filea IF "abc" = "def"',   []),
    ('filea IF date("filea") = "2005-05-05 15:25:00"',   ['filea']),
    ('filea IF date("filea") = "2005-5-5 15:25"',        ['filea']),
    ('filea IF "2005-03-05 03:00:00" = date("filea")',   []),
    ('EACH f IN filea IF f = "filea"',                   ['filea']),
    ('EACH f IN filea IF "xyz" = f',                     []),

    ('filea IF 1 != 1',           []),
    ('filea IF 3 != 1',           ['filea']),
    ('filea IF 1.0 != 1.0',       []),
    ('filea IF 3.0 != 1.0',       ['filea']),
    ('filea IF 1   != 1.0',       []),
    ('filea IF 3   != 1.0',       ['filea']),
    ('filea IF 1.0 != 1',         []),
    ('filea IF 3.0 != 1',         ['filea']),
    ('filea IF "abc" != "abc"',   []),
    ('filea IF "abc" != "def"',   ['filea']),
    ('filea IF date("filea") != "2005-05-05 15:25:00"',   []),
    ('filea IF "2005-03-05 03:00:00" != date("filea")',   ['filea']),
    ('EACH f IN filea IF f != "filea"',                   []),
    ('EACH f IN filea IF "xyz" != f',                     ['filea']),

    # Some nested expression
    ('filea IF (1=1 AND 1=1) AND NOT (NOT 1=1 OR 1=0)',         ['filea']),
    #            T   T   T   *T*  T    F   T  F   F 
    ('filea IF NOT ( (1=1 AND 1=1) AND NOT (NOT 1=1 OR 1=0) )', []),
    ('filea IF 1=1 AND (1=1 AND (1=1 AND 1=1))',                ['filea']),
    ('filea IF 1=1 AND (1=1 AND (1=1 AND 1=0))',                []),
    ('filea IF 1=0 OR (1=0 OR (1=0 OR 1=1))',                   ['filea']),

    # Functions. They are tested primarly in test_functions;
    # here we just test that the function call interface seems
    # to be working. We use functions with 0, 1 and 2 arguments.
    ('filea IF now() >= "1970-01-01 00:00:00"',        ['filea']),
    ('filea IF now() <  "1970-01-01 00:00:00"',        []),
    ('filea IF size("filea") = 100',                   ['filea']),
    ('filea IF size("filea") =  10',                   []),
    ('EACH f IN * IF size(f) = 100',                   ['filea']),
    ('EACH f IN * IF size(f) = 666',                   []),
    ('filea IF extract(date("filea"), "year") = 2005', ['filea']),
    ('filea IF extract(date("filea"), "year") = 2004', []),

    # Datetime literals
    ('filea IF date("filea") < "2004"',   []),
    ('filea IF date("filea") < "2004-08"',   []),
    ('filea IF date("filea") < "2004-08-25"',   []),
    ('filea IF date("filea") < "2004-08-25 14:26"',   []),
    ('filea IF date("filea") < "2004-08-25 14:26:47"',   []),
]


def rootdir_combinations():
    cwd = os.getcwd().replace('\\', '/')
    return [
        tutils.ROOTDIR,
        tutils.ROOTDIR + '/',
        './' + tutils.ROOTDIR,
#        './' + tutils.ROOTDIR + '/',
        os.path.abspath(tutils.ROOTDIR),
#        os.path.abspath(tutils.ROOTDIR) + '/',
    ]

def _interpret(filename, rootdir):
    files = Interpreter.interpret(filename, rootdir)
    files.sort()
    return files

def _interpret_string(string, rootdir):
    files = Interpreter.interpret_string(string, rootdir)
    files.sort()
    return files

class TestInterpreter(unittest.TestCase):

    def setUp(self):
        self.old_stdout = sys.stdout
        self.old_stderr = sys.stderr
        sys.stdout = cStringIO.StringIO()
        sys.stderr = cStringIO.StringIO()

    def tearDown(self):
        sys.stdout = self.old_stdout
        sys.stderr = self.old_stderr

    def _test_nofiles_generic(self, source_filename):
        self._check_filenames(
            tutils.get_interpreter_file(source_filename),
            [])

    def _test_allfiles_generic(self, source_filename):
        self._check_filenames(
            tutils.get_interpreter_file(source_filename),
            ALL_FILES)

    def _check_filenames(self, source_filename, expected_filenames, source_msg=None):
        if source_msg is None:
            source_msg = source_filename
        for rootdir in rootdir_combinations():
            files = _interpret(source_filename, rootdir)

            for fileobj in files:
#                self.failUnless(os.path.exists(fileobj.rel_filename),
#                    'Interpreter produced "%s", but the file was not found by os.path.exists' % fileobj.filename)
                self.failUnless(os.path.exists(fileobj.abs_filename),
                    'Interpreter produced "%s", but the file was not found by os.path.exists' % fileobj.abs_filename)

            actual_set = set([filobj.abs_filename for filobj in files])
            expected_set = set()
            for i, exp_filename in enumerate(expected_filenames):
                abspath = fsl.utils.abspath(fsl.utils.combine_rootdir(rootdir, exp_filename))
                expected_set.add(abspath)

            if expected_set != actual_set:
                missing_set = expected_set - actual_set
                extraneous_set = actual_set - expected_set

                msg = '"%s" didn\'t produce expected files (rootdir=\'%s\')' % (source_msg, rootdir)

                if missing_set:
                    missing_list = list(missing_set)
                    missing_list.sort()
                    missing_str = ', '.join(['"%s"' % s for s in missing_list])
                    msg += '\nMissing: ' + missing_str

                if extraneous_set:
                    extraneous_list = list(extraneous_set)
                    extraneous_list.sort()
                    extraneous_str = ', '.join(['"%s"' % s for s in extraneous_list])
                    msg += '\nExtraneous: ' + extraneous_str

                self.fail(msg)

    def _check_filenames_str(self, source_str, expected_filenames):
        source_list = [source_str]
        if '/' in source_str:
            source_list.append(source_str.replace('/', '\\'))

        for source_string in source_list:
            outfile = file(TMP_SOURCE_FILE, 'wt')
            outfile.write(source_string + '\n')
            outfile.close()
            self._check_filenames(TMP_SOURCE_FILE, expected_filenames, source_string)

    ### Single glob ####################################################

    def test_singleglob(self):
        for source_str, expected_files in SINGLE_GLOB_TESTS:
            self._check_filenames_str(source_str, expected_files)

    def test_singleglob_exclusion(self):
        for source_str, expected_files in SINGLE_GLOB_EXCLUSION_TESTS:
            self._check_filenames_str(source_str, expected_files)

    def test_singleglob_auto_exclusion(self):
        for source_str, expected_files in SINGLE_GLOB_TESTS:
            source_lower = source_str.lower()
            if source_lower.startswith('not ') or ' nonrec' in source_lower:
                continue
            source2 = '*\nNOT %s' % source_str
            expected2 = ALL_FILES - set(expected_files)
            self._check_filenames_str(source2, expected2)


    ### Glob list ######################################################

    def test_globlist_1(self):
        self._check_filenames_str(
            '*/file*, ALLCAPS',
            ALL_FILE_ASTERISK | set(['ALLCAPS']))

    def test_globlist_2(self):
        self._check_filenames_str(
            '*/file* NONREC, ALLCAPS',
            FIRST_LEVEL_FILE_ASTERISK | set(['ALLCAPS']) )

    def test_globlist_3(self):
        self._check_filenames_str(
            'dir1/*, */file* NONREC',
            DIR1_FILES | FIRST_LEVEL_FILE_ASTERISK)

    def test_globlist_4(self):
        self._check_filenames_str(
            'dir1/* NONREC, */file* NONREC',
            DIR1_NONREC_FILES | FIRST_LEVEL_FILE_ASTERISK)

    def test_globlist_5(self):
        self._check_filenames_str(
            'file*, ALLCAPS, dir1/filea, .hidden',
            FIRST_LEVEL_FILE_ASTERISK | set(['ALLCAPS', 'dir1/filea', '.hidden']))

    ### Glob list with IF-expression ###################################

    def test_globlist_if_1(self):
        self._check_filenames_str(
            'file* IF exists("and")',
            FIRST_LEVEL_FILE_ASTERISK)

    def test_globlist_if_2(self):
        self._check_filenames_str(
            'file*, ALLCAPS IF exists("and")',
            FIRST_LEVEL_FILE_ASTERISK | set(['ALLCAPS']))

    def test_globlist_if_3(self):
        self._check_filenames_str(
            'file* IF 1>2',
            [])

    def test_globlist_if_4(self):
        self._check_filenames_str(
            'file*, ALLCAPS IF 1>2',
            [])

    ### For each #######################################################
    # Exclusion for-each rules are tested in conjunction with a "*" rule.

    # If-expression true for all files

    def test_foreach_1(self):
        self._check_filenames_str(
            'EACH f IN * IF exists(f)',
            ALL_FILES)

    def test_foreach_not_1(self):
        self._check_filenames_str(
            '*\nNOT EACH f IN * IF exists(f)',
            [])

    def test_foreach_2(self):
        self._check_filenames_str(
            'EACH f IN file*, ALLCAPS IF exists(f)',
            FIRST_LEVEL_FILE_ASTERISK | set(['ALLCAPS']))

    def test_foreach_not_2(self):
        self._check_filenames_str(
            '*\nNOT EACH f IN file*, ALLCAPS IF exists(f)',
            ALL_FILES - (FIRST_LEVEL_FILE_ASTERISK | set(['ALLCAPS'])))

    # If-expression true for some files

    def test_foreach_3(self):
        self._check_filenames_str(
            'EACH f IN * IF size(f) >= 95',
            set(['filea', 'dir1/filea-a-a']))

    def test_foreach_not_3(self):
        self._check_filenames_str(
            '*\nNOT EACH f IN * NONREC IF size(f) >= 95',
            ALL_FILES - set(['filea']))

    def test_foreach_4(self):
        self._check_filenames_str(
            'EACH f IN file*, ALLCAPS IF size(f) = 100',
            set(['filea']))

    def test_foreach_not_4(self):
        self._check_filenames_str(
            '*\nNOT EACH f IN file*, ALLCAPS IF size(f) = 100',
            ALL_FILES - set(['filea']))

    # If-expression false for all files

    def test_foreach_5(self):
        self._check_filenames_str(
            'EACH f IN * IF NOT exists(f)',
            [])

    def test_foreach_not_5(self):
        self._check_filenames_str(
            '*\nNOT EACH f IN * IF NOT exists(f)',
            ALL_FILES)

    def test_foreach_6(self):
        self._check_filenames_str(
            'EACH f IN file*, ALLCAPS IF 1>2',
            [])

    def test_foreach_not_6(self):
        self._check_filenames_str(
            '*\nNOT EACH f IN file*, ALLCAPS IF 1>2',
            ALL_FILES)

    # No IN-section

    def test_foreach_7(self):
        self._check_filenames_str(
            'EACH f IF exists(f)',
            ALL_FILES)

    def test_foreach_not_7(self):
        self._check_filenames_str(
            '*\nNOT EACH f IF exists(f)',
            [])

    def test_foreach_8(self):
        self._check_filenames_str(
            'EACH f IF size(f) >= 95',
            set(['filea', 'dir1/filea-a-a']))

    def test_foreach_not_8(self):
        self._check_filenames_str(
            '*\nNOT EACH f IF size(f) >= 95',
            ALL_FILES - set(['filea', 'dir1/filea-a-a']))

    # Directory name in glob

    def test_foreach_9(self):
        self._check_filenames_str(
            'EACH f IN dir1 IF 1=1',
            DIR1_FILES)

    def test_foreach_not_9(self):
        self._check_filenames_str(
            '*\nNOT EACH f IN dir1 IF 1=1',
            ALL_FILES - DIR1_FILES)

    # NONREC

    def test_foreach_10(self):
        self._check_filenames_str(
            'EACH f IN dir1/* NONREC IF exists(f)',
            DIR1_NONREC_FILES)

    def test_foreach_not_10(self):
        self._check_filenames_str(
            '*\nNOT EACH f IN dir1/* NONREC IF exists(f)',
            ALL_FILES - DIR1_NONREC_FILES)

    # File name in glob

    def test_foreach_11(self):
        self._check_filenames_str(
            'EACH f IN filea IF 1=1',
            ['filea'])

    def test_foreach_not_11(self):
        self._check_filenames_str(
            '*\nNOT EACH f IN filea IF 1=1',
            ALL_FILES - set(['filea']))

    # Automatic tests using single-glob tests

    def test_foreach_globs(self):
        for source_str, expected_files in SINGLE_GLOB_TESTS:
            if source_str.lower().startswith('not '):
                continue
            source2 = 'EACH f IN %s IF 1=1' % source_str
            self._check_filenames_str(source2, expected_files)

    def test_foreach_globs_exclusion(self):
        for source_str, expected_files in SINGLE_GLOB_TESTS:
            source_lower = source_str.lower()
            if source_lower.startswith('not ') or ' nonrec' in source_lower:
                continue
            source2 = '*\nNOT EACH f IN %s IF 1=1' % source_str
            expected2 = ALL_FILES - set(expected_files)
            self._check_filenames_str(source2, expected2)


    ### Rule sequences #################################################

    def _generic_ruleseq(self, number, expected_files):
        self._check_filenames(
            tutils.get_interpreter_file('valid_ruleseq_%d' % number),
            expected_files)

    def test_ruleseq_1(self):
        self._generic_ruleseq(1, FIRST_LEVEL_FILE_ASTERISK | set(['ALLCAPS']))

    def test_ruleseq_2(self):
        self._generic_ruleseq(2, ['ALLCAPS', 'filea'])

    def test_ruleseq_3(self):
        self._generic_ruleseq(3, ['dir1/filea', 'dir xyz/aaa'])

    def test_ruleseq_4(self):
        self._generic_ruleseq(4, FIRST_LEVEL_FILE_ASTERISK | set(['.hidden', 'dir xyz/aaa']))

    def test_ruleseq_5(self):
        self._generic_ruleseq(5, FIRST_LEVEL_FILE_ASTERISK | set(['.hidden', 'dir xyz/aaa']))

    def test_ruleseq_6(self):
        self._generic_ruleseq(6, FIRST_LEVEL_FILE_ASTERISK | set(['.hidden', 'dir xyz/aaa']))

    def test_ruleseq_7(self):
        self._generic_ruleseq(7, FIRST_LEVEL_FILE_ASTERISK | set(['.hidden', 'dir xyz/aaa']))

    def test_ruleseq_8(self):
        self._generic_ruleseq(8, FIRST_LEVEL_FILE_ASTERISK | set(['.hidden', 'dir xyz/aaa']))

    def test_ruleseq_9(self):
        self._generic_ruleseq(9, FIRST_LEVEL_FILE_ASTERISK | set(['.hidden', 'dir xyz/aaa']))

    def test_ruleseq_10(self):
        self._generic_ruleseq(10, set(['.hidden', 'ALLCAPS', 'dir1/filea']))

    ### IN-blocks ######################################################

    def _generic_inblock(self, number, expected_files):
        self._check_filenames(
            tutils.get_interpreter_file('valid_inblock_%d' % number),
            expected_files)

    def test_inblock_1(self):
        self._generic_inblock(1, DIR1_FILE_ASTERISK)

    def test_inblock_2(self):
        self._generic_inblock(2, DIR1_FILES)

    def test_inblock_3(self):
        self._generic_inblock(3, ['dir1/filea'])

    def test_inblock_4(self):
        self._generic_inblock(4, [])

    def test_inblock_5(self):
        self._generic_inblock(5, [])

    def test_inblock_6(self):
        self._generic_inblock(6, [])

    # Nested IN-blocks

    def test_inblock_7(self):
        self._generic_inblock(7, ['dir1/dir2/asdf'])

    def test_inblock_8(self):
        self._generic_inblock(8, DIR1_DIR2_FILES | set(['dir1/filea']))

    def test_inblock_9(self):
        self._generic_inblock(9, DIR1_DIR2_DIR3_FILES)

    def test_inblock_10(self):
        self._generic_inblock(10, [])

    ### IF-blocks ######################################################

    def _generic_ifblock(self, number, expected_files):
        self._check_filenames(
            tutils.get_interpreter_file('valid_ifblock_%d' % number),
            expected_files)

    def test_ifblock_1(self):
        self._generic_ifblock(1, ['filea', 'filed'])

    def test_ifblock_2(self):
        self._generic_ifblock(2, [])

    def test_ifblock_3(self):
        self._generic_ifblock(3, ['filea', 'filed', 'dir1/filea'])

    ### Inclusion/exclusion combinations ###############################

    def _generic_incl_excl(self, number, expected_files):
        self._check_filenames(
            tutils.get_interpreter_file('valid_incl_excl_%d' % number),
            expected_files)

    def test_incl_excl_1(self):
        self._generic_incl_excl(1, [])

    def test_incl_excl_2(self):
        self._generic_incl_excl(2, ALL_FILES)

    def test_incl_excl_3(self):
        self._generic_incl_excl(3, ALL_FILES)

    def test_incl_excl_4(self):
        self._generic_incl_excl(4, [])
    
    # A=filea B=fileb.txt C=dir1/filea-a-a D=ALLCAPS 

    def test_incl_excl_5(self):
        self._generic_incl_excl(5, ['fileb.txt', 'dir1/filea-a-a'])

    def test_incl_excl_6(self):
        self._generic_incl_excl(6, ['fileb.txt', 'dir1/filea-a-a'])

    def test_incl_excl_7(self):
        self._generic_incl_excl(7, ['filea'])

    def test_incl_excl_8(self):
        self._generic_incl_excl(8, [])

    def test_incl_excl_9(self):
        self._generic_incl_excl(9, ['filea', 'fileb.txt', 'dir1/filea-a-a'])

    def test_incl_excl_10(self):
        self._generic_incl_excl(10, ['fileb.txt', 'dir1/filea-a-a', 'ALLCAPS'])

    # Test * * and such

    def test_incl_excl_11(self):
        self._generic_incl_excl(11, ALL_FILES)

    def test_incl_excl_12(self):
        self._generic_incl_excl(12, ALL_FILES)

    # Misc

    def test_incl_excl_13(self):
        """Test IN-blocks."""
        self._generic_incl_excl(13, DIR1_NONREC_FILES)

    def test_incl_excl_14(self):
        self._generic_incl_excl(14, [])

    ### Expressions ####################################################

    def test_expressions(self):
        for source_str, expected_files in EXPRESSION_TESTS:
            self._check_filenames_str(source_str, expected_files)

    ### Hooks ##########################################################

    def _interpret_hook(self, source_filename, source_msg=None):
        if source_msg is None:
            source_msg = source_filename

        for rootdir in rootdir_combinations():
            hook_handler = TesterHookHandler.TesterHookHandler()
            hook_handler.source_msg = source_msg
            fileobjs = Interpreter.interpret(source_filename, rootdir, hook_handler)

            fileset = set([f.abs_filename for f in fileobjs])
            hook_fileset = hook_handler.include_fileset

            if fileset != hook_fileset:
                msg = '"%s": file list returned by interpreter differs from hook calls' % source_msg
                diff_set = fileset.symmetric_difference(hook_fileset)
                diff_list = list(diff_set)
                diff_list.sort()
                diff_str = ', '.join(['"%s"' % s for s in diff_list])
                msg += '\nDifference: ' + diff_str
                self.fail(msg)

            if fileset == ALL_FILES and (not hook_handler.scan_directory_dict):
                self.fail('"%s": files were produced, but scan hooks were not called' % source_msg)

            for dirname, (starts, ends) in hook_handler.scan_directory_dict.items():
                self.assertEqual(starts, ends,
                    '"%s": Unequal number of start-scan (%s) and end-scan (%s) hook calls (dirname="%s")' % (source_msg, starts, ends, dirname))

    def _interpret_hook_str(self, source_str):
        outfile = file(TMP_SOURCE_FILE, 'wt')
        outfile.write(source_str + '\n')
        outfile.close()
        self._interpret_hook(TMP_SOURCE_FILE, source_str)

    def test_include_hook_only_once(self):
        """Test that (include) hooks are called only once."""
        self._interpret_hook_str('*\n*')

    def test_exclude_hook_only_once(self):
        """Test that (exclude) hooks are called only once."""
        self._interpret_hook_str('*\nNOT *\n*\nNOT *')

    def test_hook_single_globs(self):
        for source_str, expected_files in SINGLE_GLOB_TESTS:
            self._interpret_hook_str(source_str)

    def test_hook_single_glob_exclusions(self):
        for source_str, expected_files in SINGLE_GLOB_EXCLUSION_TESTS:
            self._interpret_hook_str(source_str)

    def test_hook_files(self):
        for filename in tutils.get_interpreter_files('valid*'):
            self._interpret_hook(filename)

    ### Short circuit ##################################################

    def _generic_short_circuit(self, source_filename):
        """Test that directory scanning is not done when not necessary."""
        source_filename = tutils.get_interpreter_file(source_filename)
        for rootdir in rootdir_combinations():
            hook_handler = TesterHookHandler.TesterHookHandler()
            files = Interpreter.interpret(source_filename, rootdir, hook_handler)

            if hook_handler.scan_directory_dict:
                dirs = [d for d in hook_handler.scan_directory_dict.keys()]
                dirs_str = ', '.join(dirs)
                self.fail('"%s": expected no directory scans, but these were scanned: %s' % (source_filename, dirs_str))

            self.failIf(files, '"%s": expected no files, got %d files' % (source_filename, len(files)))

    def test_short_circuit_1(self):
        self._generic_short_circuit('valid_short_circuit_1')

    def test_short_circuit_2(self):
        self._generic_short_circuit('valid_short_circuit_2')

#    def test_short_circuit_3(self):
#        self._generic_short_circuit('valid_short_circuit_3')

    def test_short_circuit_4(self):
        self._generic_short_circuit('valid_short_circuit_4')

#    def test_short_circuit_5(self):
#        self._generic_short_circuit('valid_short_circuit_5')

#    def test_short_circuit_6(self):
#        self._generic_short_circuit('valid_short_circuit_6')

    def test_short_circuit_7(self):
        self._generic_short_circuit('valid_short_circuit_7')

#    def test_short_circuit_8(self):
#        self._generic_short_circuit('valid_short_circuit_8')

    ### All files ######################################################

    def test_allfiles_1(self):
        self._test_allfiles_generic('valid_allfiles_1')

    def test_allfiles_2(self):
        self._test_allfiles_generic('valid_allfiles_2')

    def test_allfiles_3(self):
        self._test_allfiles_generic('valid_allfiles_3')

    def test_allfiles_4(self):
        self._test_allfiles_generic('valid_allfiles_4')

    def test_allfiles_5(self):
        self._test_allfiles_generic('valid_allfiles_5')

    def test_allfiles_6(self):
        self._test_allfiles_generic('valid_allfiles_6')

    # No files

    def test_nofiles_1(self):
        self._test_nofiles_generic('valid_nofiles_1')

    def test_nofiles_2(self):
        self._test_nofiles_generic('valid_nofiles_2')

    def test_nofiles_3(self):
        self._test_nofiles_generic('valid_nofiles_3')

    def test_nofiles_4(self):
        self._test_nofiles_generic('valid_nofiles_4')

    ### String interpretation ##########################################

    def test_interpret_string(self):
        for source_str, expected_files in SINGLE_GLOB_TESTS:
            expected_set = set(expected_files)
            for rootdir in rootdir_combinations():
                string_result = set(_interpret_string(source_str, rootdir))

                outfile = file(TMP_SOURCE_FILE, 'wt')
                outfile.write(source_str + '\n')
                outfile.close()
                file_result = set(_interpret(TMP_SOURCE_FILE, rootdir))

                self.failUnless(string_result == file_result,
                    'String interpreting gives incorrect result ("%s")' % source_str)

    ### Absolute paths, single glob ####################################

    def test_absolute_single_glob(self):
        root_prefix = os.getcwd().replace('\\', '/')
        if not root_prefix.endswith('/'):
            root_prefix += '/'
        root_prefix += tutils.ROOTDIR + '/'

        for source_str, expected_files in SINGLE_GLOB_TESTS:
            if source_str.startswith('*') or source_str.startswith('"'):
                continue
            source_str = root_prefix + source_str
            self._check_filenames_str(source_str, expected_files)

def main(args):
#    print rootdir_combinations()
    unittest.main()

#    suite = unittest.defaultTestLoader.loadTestsFromName('test_interpreter.TestInterpreter.test_absolute_single_glob')
#    unittest.TextTestRunner().run(suite)

if __name__=='__main__':
    import sys
    main(sys.argv[1:])
